from .helper import PillowTestCase, hopper


class TestImagePoint(PillowTestCase):
    def test_sanity(self):
        im = hopper()

        self.assertRaises(ValueError, im.point, list(range(256)))
        im.point(list(range(256)) * 3)
        im.point(lambda x: x)

        im = im.convert("I")
        self.assertRaises(ValueError, im.point, list(range(256)))
        im.point(lambda x: x * 1)
        im.point(lambda x: x + 1)
        im.point(lambda x: x * 1 + 1)
        self.assertRaises(TypeError, im.point, lambda x: x - 1)
        self.assertRaises(TypeError, im.point, lambda x: x / 1)

    def test_16bit_lut(self):
        """ Tests for 16 bit -> 8 bit lut for converting I->L images
            see https://github.com/python-pillow/Pillow/issues/440
            """
        im = hopper("I")
        im.point(list(range(256)) * 256, "L")

    def test_f_lut(self):
        """ Tests for floating point lut of 8bit gray image """
        im = hopper("L")
        lut = [0.5 * float(x) for x in range(256)]

        out = im.point(lut, "F")

        int_lut = [x // 2 for x in range(256)]
        self.assert_image_equal(out.convert("L"), im.point(int_lut, "L"))

    def test_f_mode(self):
        im = hopper("F")
        self.assertRaises(ValueError, im.point, None)



from PIL import BdfFontFile, FontFile

from .helper import PillowTestCase

filename = "Tests/images/courB08.bdf"


class TestFontBdf(PillowTestCase):
    def test_sanity(self):

        with open(filename, "rb") as test_file:
            font = BdfFontFile.BdfFontFile(test_file)

        self.assertIsInstance(font, FontFile.FontFile)
        self.assertEqual(len([_f for _f in font.glyph if _f]), 190)

    def test_invalid_file(self):
        with open("Tests/images/flower.jpg", "rb") as fp:
            self.assertRaises(SyntaxError, BdfFontFile.BdfFontFile, fp)



import io

from PIL import EpsImagePlugin, Image

from .helper import PillowTestCase, hopper, unittest

HAS_GHOSTSCRIPT = EpsImagePlugin.has_ghostscript()

# Our two EPS test files (they are identical except for their bounding boxes)
file1 = "Tests/images/zero_bb.eps"
file2 = "Tests/images/non_zero_bb.eps"

# Due to palletization, we'll need to convert these to RGB after load
file1_compare = "Tests/images/zero_bb.png"
file1_compare_scale2 = "Tests/images/zero_bb_scale2.png"

file2_compare = "Tests/images/non_zero_bb.png"
file2_compare_scale2 = "Tests/images/non_zero_bb_scale2.png"

# EPS test files with binary preview
file3 = "Tests/images/binary_preview_map.eps"


class TestFileEps(PillowTestCase):
    @unittest.skipUnless(HAS_GHOSTSCRIPT, "Ghostscript not available")
    def test_sanity(self):
        # Regular scale
        image1 = Image.open(file1)
        image1.load()
        self.assertEqual(image1.mode, "RGB")
        self.assertEqual(image1.size, (460, 352))
        self.assertEqual(image1.format, "EPS")

        image2 = Image.open(file2)
        image2.load()
        self.assertEqual(image2.mode, "RGB")
        self.assertEqual(image2.size, (360, 252))
        self.assertEqual(image2.format, "EPS")

        # Double scale
        image1_scale2 = Image.open(file1)
        image1_scale2.load(scale=2)
        self.assertEqual(image1_scale2.mode, "RGB")
        self.assertEqual(image1_scale2.size, (920, 704))
        self.assertEqual(image1_scale2.format, "EPS")

        image2_scale2 = Image.open(file2)
        image2_scale2.load(scale=2)
        self.assertEqual(image2_scale2.mode, "RGB")
        self.assertEqual(image2_scale2.size, (720, 504))
        self.assertEqual(image2_scale2.format, "EPS")

    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(SyntaxError, EpsImagePlugin.EpsImageFile, invalid_file)

    @unittest.skipUnless(HAS_GHOSTSCRIPT, "Ghostscript not available")
    def test_cmyk(self):
        cmyk_image = Image.open("Tests/images/pil_sample_cmyk.eps")

        self.assertEqual(cmyk_image.mode, "CMYK")
        self.assertEqual(cmyk_image.size, (100, 100))
        self.assertEqual(cmyk_image.format, "EPS")

        cmyk_image.load()
        self.assertEqual(cmyk_image.mode, "RGB")

        if "jpeg_decoder" in dir(Image.core):
            target = Image.open("Tests/images/pil_sample_rgb.jpg")
            self.assert_image_similar(cmyk_image, target, 10)

    @unittest.skipUnless(HAS_GHOSTSCRIPT, "Ghostscript not available")
    def test_showpage(self):
        # See https://github.com/python-pillow/Pillow/issues/2615
        plot_image = Image.open("Tests/images/reqd_showpage.eps")
        target = Image.open("Tests/images/reqd_showpage.png")

        # should not crash/hang
        plot_image.load()
        #  fonts could be slightly different
        self.assert_image_similar(plot_image, target, 6)

    @unittest.skipUnless(HAS_GHOSTSCRIPT, "Ghostscript not available")
    def test_file_object(self):
        # issue 479
        image1 = Image.open(file1)
        with open(self.tempfile("temp_file.eps"), "wb") as fh:
            image1.save(fh, "EPS")

    @unittest.skipUnless(HAS_GHOSTSCRIPT, "Ghostscript not available")
    def test_iobase_object(self):
        # issue 479
        image1 = Image.open(file1)
        with io.open(self.tempfile("temp_iobase.eps"), "wb") as fh:
            image1.save(fh, "EPS")

    @unittest.skipUnless(HAS_GHOSTSCRIPT, "Ghostscript not available")
    def test_bytesio_object(self):
        with open(file1, "rb") as f:
            img_bytes = io.BytesIO(f.read())

        img = Image.open(img_bytes)
        img.load()

        image1_scale1_compare = Image.open(file1_compare).convert("RGB")
        image1_scale1_compare.load()
        self.assert_image_similar(img, image1_scale1_compare, 5)

    def test_image_mode_not_supported(self):
        im = hopper("RGBA")
        tmpfile = self.tempfile("temp.eps")
        self.assertRaises(ValueError, im.save, tmpfile)

    @unittest.skipUnless(HAS_GHOSTSCRIPT, "Ghostscript not available")
    def test_render_scale1(self):
        # We need png support for these render test
        codecs = dir(Image.core)
        if "zip_encoder" not in codecs or "zip_decoder" not in codecs:
            self.skipTest("zip/deflate support not available")

        # Zero bounding box
        image1_scale1 = Image.open(file1)
        image1_scale1.load()
        image1_scale1_compare = Image.open(file1_compare).convert("RGB")
        image1_scale1_compare.load()
        self.assert_image_similar(image1_scale1, image1_scale1_compare, 5)

        # Non-Zero bounding box
        image2_scale1 = Image.open(file2)
        image2_scale1.load()
        image2_scale1_compare = Image.open(file2_compare).convert("RGB")
        image2_scale1_compare.load()
        self.assert_image_similar(image2_scale1, image2_scale1_compare, 10)

    @unittest.skipUnless(HAS_GHOSTSCRIPT, "Ghostscript not available")
    def test_render_scale2(self):
        # We need png support for these render test
        codecs = dir(Image.core)
        if "zip_encoder" not in codecs or "zip_decoder" not in codecs:
            self.skipTest("zip/deflate support not available")

        # Zero bounding box
        image1_scale2 = Image.open(file1)
        image1_scale2.load(scale=2)
        image1_scale2_compare = Image.open(file1_compare_scale2).convert("RGB")
        image1_scale2_compare.load()
        self.assert_image_similar(image1_scale2, image1_scale2_compare, 5)

        # Non-Zero bounding box
        image2_scale2 = Image.open(file2)
        image2_scale2.load(scale=2)
        image2_scale2_compare = Image.open(file2_compare_scale2).convert("RGB")
        image2_scale2_compare.load()
        self.assert_image_similar(image2_scale2, image2_scale2_compare, 10)

    @unittest.skipUnless(HAS_GHOSTSCRIPT, "Ghostscript not available")
    def test_resize(self):
        # Arrange
        image1 = Image.open(file1)
        image2 = Image.open(file2)
        image3 = Image.open("Tests/images/illu10_preview.eps")
        new_size = (100, 100)

        # Act
        image1 = image1.resize(new_size)
        image2 = image2.resize(new_size)
        image3 = image3.resize(new_size)

        # Assert
        self.assertEqual(image1.size, new_size)
        self.assertEqual(image2.size, new_size)
        self.assertEqual(image3.size, new_size)

    @unittest.skipUnless(HAS_GHOSTSCRIPT, "Ghostscript not available")
    def test_thumbnail(self):
        # Issue #619
        # Arrange
        image1 = Image.open(file1)
        image2 = Image.open(file2)
        new_size = (100, 100)

        # Act
        image1.thumbnail(new_size)
        image2.thumbnail(new_size)

        # Assert
        self.assertEqual(max(image1.size), max(new_size))
        self.assertEqual(max(image2.size), max(new_size))

    def test_read_binary_preview(self):
        # Issue 302
        # open image with binary preview
        Image.open(file3)

    def _test_readline(self, t, ending):
        ending = "Failure with line ending: %s" % (
            "".join("%s" % ord(s) for s in ending)
        )
        self.assertEqual(t.readline().strip("\r\n"), "something", ending)
        self.assertEqual(t.readline().strip("\r\n"), "else", ending)
        self.assertEqual(t.readline().strip("\r\n"), "baz", ending)
        self.assertEqual(t.readline().strip("\r\n"), "bif", ending)

    def _test_readline_io_psfile(self, test_string, ending):
        f = io.BytesIO(test_string.encode("latin-1"))
        t = EpsImagePlugin.PSFile(f)
        self._test_readline(t, ending)

    def _test_readline_file_psfile(self, test_string, ending):
        f = self.tempfile("temp.txt")
        with open(f, "wb") as w:
            w.write(test_string.encode("latin-1"))

        with open(f, "rb") as r:
            t = EpsImagePlugin.PSFile(r)
            self._test_readline(t, ending)

    def test_readline(self):
        # check all the freaking line endings possible from the spec
        # test_string = u'something\r\nelse\n\rbaz\rbif\n'
        line_endings = ["\r\n", "\n", "\n\r", "\r"]
        strings = ["something", "else", "baz", "bif"]

        for ending in line_endings:
            s = ending.join(strings)
            self._test_readline_io_psfile(s, ending)
            self._test_readline_file_psfile(s, ending)

    def test_open_eps(self):
        # https://github.com/python-pillow/Pillow/issues/1104
        # Arrange
        FILES = [
            "Tests/images/illu10_no_preview.eps",
            "Tests/images/illu10_preview.eps",
            "Tests/images/illuCS6_no_preview.eps",
            "Tests/images/illuCS6_preview.eps",
        ]

        # Act / Assert
        for filename in FILES:
            img = Image.open(filename)
            self.assertEqual(img.mode, "RGB")

    @unittest.skipUnless(HAS_GHOSTSCRIPT, "Ghostscript not available")
    def test_emptyline(self):
        # Test file includes an empty line in the header data
        emptyline_file = "Tests/images/zero_bb_emptyline.eps"

        image = Image.open(emptyline_file)
        image.load()
        self.assertEqual(image.mode, "RGB")
        self.assertEqual(image.size, (460, 352))
        self.assertEqual(image.format, "EPS")



from PIL import Image, ImImagePlugin

from .helper import PillowTestCase, hopper

# sample im
TEST_IM = "Tests/images/hopper.im"


class TestFileIm(PillowTestCase):
    def test_sanity(self):
        im = Image.open(TEST_IM)
        im.load()
        self.assertEqual(im.mode, "RGB")
        self.assertEqual(im.size, (128, 128))
        self.assertEqual(im.format, "IM")

    def test_unclosed_file(self):
        def open():
            im = Image.open(TEST_IM)
            im.load()

        self.assert_warning(None, open)

    def test_tell(self):
        # Arrange
        im = Image.open(TEST_IM)

        # Act
        frame = im.tell()

        # Assert
        self.assertEqual(frame, 0)

    def test_n_frames(self):
        im = Image.open(TEST_IM)
        self.assertEqual(im.n_frames, 1)
        self.assertFalse(im.is_animated)

    def test_eoferror(self):
        im = Image.open(TEST_IM)
        n_frames = im.n_frames

        # Test seeking past the last frame
        self.assertRaises(EOFError, im.seek, n_frames)
        self.assertLess(im.tell(), n_frames)

        # Test that seeking to the last frame does not raise an error
        im.seek(n_frames - 1)

    def test_roundtrip(self):
        for mode in ["RGB", "P", "PA"]:
            out = self.tempfile("temp.im")
            im = hopper(mode)
            im.save(out)
            reread = Image.open(out)

            self.assert_image_equal(reread, im)

    def test_save_unsupported_mode(self):
        out = self.tempfile("temp.im")
        im = hopper("HSV")
        self.assertRaises(ValueError, im.save, out)

    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(SyntaxError, ImImagePlugin.ImImageFile, invalid_file)

    def test_number(self):
        self.assertEqual(1.2, ImImagePlugin.number("1.2"))



import io

from PIL import IcoImagePlugin, Image, ImageDraw

from .helper import PillowTestCase, hopper

TEST_ICO_FILE = "Tests/images/hopper.ico"


class TestFileIco(PillowTestCase):
    def test_sanity(self):
        im = Image.open(TEST_ICO_FILE)
        im.load()
        self.assertEqual(im.mode, "RGBA")
        self.assertEqual(im.size, (16, 16))
        self.assertEqual(im.format, "ICO")
        self.assertEqual(im.get_format_mimetype(), "image/x-icon")

    def test_invalid_file(self):
        with open("Tests/images/flower.jpg", "rb") as fp:
            self.assertRaises(SyntaxError, IcoImagePlugin.IcoImageFile, fp)

    def test_save_to_bytes(self):
        output = io.BytesIO()
        im = hopper()
        im.save(output, "ico", sizes=[(32, 32), (64, 64)])

        # the default image
        output.seek(0)
        reloaded = Image.open(output)
        self.assertEqual(reloaded.info["sizes"], {(32, 32), (64, 64)})

        self.assertEqual(im.mode, reloaded.mode)
        self.assertEqual((64, 64), reloaded.size)
        self.assertEqual(reloaded.format, "ICO")
        self.assert_image_equal(reloaded, hopper().resize((64, 64), Image.LANCZOS))

        # the other one
        output.seek(0)
        reloaded = Image.open(output)
        reloaded.size = (32, 32)

        self.assertEqual(im.mode, reloaded.mode)
        self.assertEqual((32, 32), reloaded.size)
        self.assertEqual(reloaded.format, "ICO")
        self.assert_image_equal(reloaded, hopper().resize((32, 32), Image.LANCZOS))

    def test_incorrect_size(self):
        im = Image.open(TEST_ICO_FILE)
        with self.assertRaises(ValueError):
            im.size = (1, 1)

    def test_save_256x256(self):
        """Issue #2264 https://github.com/python-pillow/Pillow/issues/2264"""
        # Arrange
        im = Image.open("Tests/images/hopper_256x256.ico")
        outfile = self.tempfile("temp_saved_hopper_256x256.ico")

        # Act
        im.save(outfile)
        im_saved = Image.open(outfile)

        # Assert
        self.assertEqual(im_saved.size, (256, 256))

    def test_only_save_relevant_sizes(self):
        """Issue #2266 https://github.com/python-pillow/Pillow/issues/2266
        Should save in 16x16, 24x24, 32x32, 48x48 sizes
        and not in 16x16, 24x24, 32x32, 48x48, 48x48, 48x48, 48x48 sizes
        """
        # Arrange
        im = Image.open("Tests/images/python.ico")  # 16x16, 32x32, 48x48
        outfile = self.tempfile("temp_saved_python.ico")

        # Act
        im.save(outfile)
        im_saved = Image.open(outfile)

        # Assert
        self.assertEqual(
            im_saved.info["sizes"], {(16, 16), (24, 24), (32, 32), (48, 48)}
        )

    def test_unexpected_size(self):
        # This image has been manually hexedited to state that it is 16x32
        # while the image within is still 16x16
        im = self.assert_warning(
            UserWarning, Image.open, "Tests/images/hopper_unexpected.ico"
        )
        self.assertEqual(im.size, (16, 16))

    def test_draw_reloaded(self):
        im = Image.open(TEST_ICO_FILE)
        outfile = self.tempfile("temp_saved_hopper_draw.ico")

        draw = ImageDraw.Draw(im)
        draw.line((0, 0) + im.size, "#f00")
        im.save(outfile)

        im = Image.open(outfile)
        im.save("Tests/images/hopper_draw.ico")
        reloaded = Image.open("Tests/images/hopper_draw.ico")
        self.assert_image_equal(im, reloaded)



from PIL import Image

from .helper import PillowTestCase

try:
    from PIL import _webp

    HAVE_WEBP = True
except ImportError:
    HAVE_WEBP = False


class TestFileWebpAnimation(PillowTestCase):
    def setUp(self):
        if not HAVE_WEBP:
            self.skipTest("WebP support not installed")
            return

        if not _webp.HAVE_WEBPANIM:
            self.skipTest(
                "WebP library does not contain animation support, "
                "not testing animation"
            )

    def test_n_frames(self):
        """
        Ensure that WebP format sets n_frames and is_animated
        attributes correctly.
        """

        im = Image.open("Tests/images/hopper.webp")
        self.assertEqual(im.n_frames, 1)
        self.assertFalse(im.is_animated)

        im = Image.open("Tests/images/iss634.webp")
        self.assertEqual(im.n_frames, 42)
        self.assertTrue(im.is_animated)

    def test_write_animation_L(self):
        """
        Convert an animated GIF to animated WebP, then compare the
        frame count, and first and last frames to ensure they're
        visually similar.
        """

        orig = Image.open("Tests/images/iss634.gif")
        self.assertGreater(orig.n_frames, 1)

        temp_file = self.tempfile("temp.webp")
        orig.save(temp_file, save_all=True)
        im = Image.open(temp_file)
        self.assertEqual(im.n_frames, orig.n_frames)

        # Compare first and last frames to the original animated GIF
        orig.load()
        im.load()
        self.assert_image_similar(im, orig.convert("RGBA"), 25.0)
        orig.seek(orig.n_frames - 1)
        im.seek(im.n_frames - 1)
        orig.load()
        im.load()
        self.assert_image_similar(im, orig.convert("RGBA"), 25.0)

    def test_write_animation_RGB(self):
        """
        Write an animated WebP from RGB frames, and ensure the frames
        are visually similar to the originals.
        """

        def check(temp_file):
            im = Image.open(temp_file)
            self.assertEqual(im.n_frames, 2)

            # Compare first frame to original
            im.load()
            self.assert_image_equal(im, frame1.convert("RGBA"))

            # Compare second frame to original
            im.seek(1)
            im.load()
            self.assert_image_equal(im, frame2.convert("RGBA"))

        frame1 = Image.open("Tests/images/anim_frame1.webp")
        frame2 = Image.open("Tests/images/anim_frame2.webp")

        temp_file1 = self.tempfile("temp.webp")
        frame1.copy().save(
            temp_file1, save_all=True, append_images=[frame2], lossless=True
        )
        check(temp_file1)

        # Tests appending using a generator
        def imGenerator(ims):
            for im in ims:
                yield im

        temp_file2 = self.tempfile("temp_generator.webp")
        frame1.copy().save(
            temp_file2,
            save_all=True,
            append_images=imGenerator([frame2]),
            lossless=True,
        )
        check(temp_file2)

    def test_timestamp_and_duration(self):
        """
        Try passing a list of durations, and make sure the encoded
        timestamps and durations are correct.
        """

        durations = [0, 10, 20, 30, 40]
        temp_file = self.tempfile("temp.webp")
        frame1 = Image.open("Tests/images/anim_frame1.webp")
        frame2 = Image.open("Tests/images/anim_frame2.webp")
        frame1.save(
            temp_file,
            save_all=True,
            append_images=[frame2, frame1, frame2, frame1],
            duration=durations,
        )

        im = Image.open(temp_file)
        self.assertEqual(im.n_frames, 5)
        self.assertTrue(im.is_animated)

        # Check that timestamps and durations match original values specified
        ts = 0
        for frame in range(im.n_frames):
            im.seek(frame)
            im.load()
            self.assertEqual(im.info["duration"], durations[frame])
            self.assertEqual(im.info["timestamp"], ts)
            ts += durations[frame]

    def test_seeking(self):
        """
        Create an animated WebP file, and then try seeking through
        frames in reverse-order, verifying the timestamps and durations
        are correct.
        """

        dur = 33
        temp_file = self.tempfile("temp.webp")
        frame1 = Image.open("Tests/images/anim_frame1.webp")
        frame2 = Image.open("Tests/images/anim_frame2.webp")
        frame1.save(
            temp_file,
            save_all=True,
            append_images=[frame2, frame1, frame2, frame1],
            duration=dur,
        )

        im = Image.open(temp_file)
        self.assertEqual(im.n_frames, 5)
        self.assertTrue(im.is_animated)

        # Traverse frames in reverse, checking timestamps and durations
        ts = dur * (im.n_frames - 1)
        for frame in reversed(range(im.n_frames)):
            im.seek(frame)
            im.load()
            self.assertEqual(im.info["duration"], dur)
            self.assertEqual(im.info["timestamp"], ts)
            ts -= dur



from PIL import _util

from .helper import PillowTestCase, unittest


class TestUtil(PillowTestCase):
    def test_is_string_type(self):
        # Arrange
        color = "red"

        # Act
        it_is = _util.isStringType(color)

        # Assert
        self.assertTrue(it_is)

    def test_is_not_string_type(self):
        # Arrange
        color = (255, 0, 0)

        # Act
        it_is_not = _util.isStringType(color)

        # Assert
        self.assertFalse(it_is_not)

    def test_is_path(self):
        # Arrange
        fp = "filename.ext"

        # Act
        it_is = _util.isPath(fp)

        # Assert
        self.assertTrue(it_is)

    @unittest.skipIf(not _util.py36, "os.path support for Paths added in 3.6")
    def test_path_obj_is_path(self):
        # Arrange
        from pathlib import Path

        test_path = Path("filename.ext")

        # Act
        it_is = _util.isPath(test_path)

        # Assert
        self.assertTrue(it_is)

    def test_is_not_path(self):
        # Arrange
        filename = self.tempfile("temp.ext")
        fp = open(filename, "w").close()

        # Act
        it_is_not = _util.isPath(fp)

        # Assert
        self.assertFalse(it_is_not)

    def test_is_directory(self):
        # Arrange
        directory = "Tests"

        # Act
        it_is = _util.isDirectory(directory)

        # Assert
        self.assertTrue(it_is)

    def test_is_not_directory(self):
        # Arrange
        text = "abc"

        # Act
        it_is_not = _util.isDirectory(text)

        # Assert
        self.assertFalse(it_is_not)

    def test_deferred_error(self):
        # Arrange

        # Act
        thing = _util.deferred_error(ValueError("Some error text"))

        # Assert
        self.assertRaises(ValueError, lambda: thing.some_attr)



import tempfile
from io import BytesIO

from PIL import Image, ImageSequence, SpiderImagePlugin

from .helper import PillowTestCase, hopper

TEST_FILE = "Tests/images/hopper.spider"


class TestImageSpider(PillowTestCase):
    def test_sanity(self):
        im = Image.open(TEST_FILE)
        im.load()
        self.assertEqual(im.mode, "F")
        self.assertEqual(im.size, (128, 128))
        self.assertEqual(im.format, "SPIDER")

    def test_unclosed_file(self):
        def open():
            im = Image.open(TEST_FILE)
            im.load()

        self.assert_warning(None, open)

    def test_save(self):
        # Arrange
        temp = self.tempfile("temp.spider")
        im = hopper()

        # Act
        im.save(temp, "SPIDER")

        # Assert
        im2 = Image.open(temp)
        self.assertEqual(im2.mode, "F")
        self.assertEqual(im2.size, (128, 128))
        self.assertEqual(im2.format, "SPIDER")

    def test_tempfile(self):
        # Arrange
        im = hopper()

        # Act
        with tempfile.TemporaryFile() as fp:
            im.save(fp, "SPIDER")

            # Assert
            fp.seek(0)
            reloaded = Image.open(fp)
            self.assertEqual(reloaded.mode, "F")
            self.assertEqual(reloaded.size, (128, 128))
            self.assertEqual(reloaded.format, "SPIDER")

    def test_isSpiderImage(self):
        self.assertTrue(SpiderImagePlugin.isSpiderImage(TEST_FILE))

    def test_tell(self):
        # Arrange
        im = Image.open(TEST_FILE)

        # Act
        index = im.tell()

        # Assert
        self.assertEqual(index, 0)

    def test_n_frames(self):
        im = Image.open(TEST_FILE)
        self.assertEqual(im.n_frames, 1)
        self.assertFalse(im.is_animated)

    def test_loadImageSeries(self):
        # Arrange
        not_spider_file = "Tests/images/hopper.ppm"
        file_list = [TEST_FILE, not_spider_file, "path/not_found.ext"]

        # Act
        img_list = SpiderImagePlugin.loadImageSeries(file_list)

        # Assert
        self.assertEqual(len(img_list), 1)
        self.assertIsInstance(img_list[0], Image.Image)
        self.assertEqual(img_list[0].size, (128, 128))

    def test_loadImageSeries_no_input(self):
        # Arrange
        file_list = None

        # Act
        img_list = SpiderImagePlugin.loadImageSeries(file_list)

        # Assert
        self.assertIsNone(img_list)

    def test_isInt_not_a_number(self):
        # Arrange
        not_a_number = "a"

        # Act
        ret = SpiderImagePlugin.isInt(not_a_number)

        # Assert
        self.assertEqual(ret, 0)

    def test_invalid_file(self):
        invalid_file = "Tests/images/invalid.spider"

        self.assertRaises(IOError, Image.open, invalid_file)

    def test_nonstack_file(self):
        im = Image.open(TEST_FILE)

        self.assertRaises(EOFError, im.seek, 0)

    def test_nonstack_dos(self):
        im = Image.open(TEST_FILE)
        for i, frame in enumerate(ImageSequence.Iterator(im)):
            if i > 1:
                self.fail("Non-stack DOS file test failed")

    # for issue #4093
    def test_odd_size(self):
        data = BytesIO()
        width = 100
        im = Image.new("F", (width, 64))
        im.save(data, format="SPIDER")

        data.seek(0)
        im2 = Image.open(data)
        self.assert_image_equal(im, im2)



from __future__ import unicode_literals

import io

from PIL import features

from .helper import PillowTestCase, unittest

try:
    from PIL import _webp

    HAVE_WEBP = True
except ImportError:
    HAVE_WEBP = False


class TestFeatures(PillowTestCase):
    def test_check(self):
        # Check the correctness of the convenience function
        for module in features.modules:
            self.assertEqual(features.check_module(module), features.check(module))
        for codec in features.codecs:
            self.assertEqual(features.check_codec(codec), features.check(codec))
        for feature in features.features:
            self.assertEqual(features.check_feature(feature), features.check(feature))

    @unittest.skipUnless(HAVE_WEBP, "WebP not available")
    def test_webp_transparency(self):
        self.assertEqual(
            features.check("transp_webp"), not _webp.WebPDecoderBuggyAlpha()
        )
        self.assertEqual(features.check("transp_webp"), _webp.HAVE_TRANSPARENCY)

    @unittest.skipUnless(HAVE_WEBP, "WebP not available")
    def test_webp_mux(self):
        self.assertEqual(features.check("webp_mux"), _webp.HAVE_WEBPMUX)

    @unittest.skipUnless(HAVE_WEBP, "WebP not available")
    def test_webp_anim(self):
        self.assertEqual(features.check("webp_anim"), _webp.HAVE_WEBPANIM)

    def test_check_modules(self):
        for feature in features.modules:
            self.assertIn(features.check_module(feature), [True, False])
        for feature in features.codecs:
            self.assertIn(features.check_codec(feature), [True, False])

    def test_supported_modules(self):
        self.assertIsInstance(features.get_supported_modules(), list)
        self.assertIsInstance(features.get_supported_codecs(), list)
        self.assertIsInstance(features.get_supported_features(), list)
        self.assertIsInstance(features.get_supported(), list)

    def test_unsupported_codec(self):
        # Arrange
        codec = "unsupported_codec"
        # Act / Assert
        self.assertRaises(ValueError, features.check_codec, codec)

    def test_unsupported_module(self):
        # Arrange
        module = "unsupported_module"
        # Act / Assert
        self.assertRaises(ValueError, features.check_module, module)

    def test_pilinfo(self):
        buf = io.StringIO()
        features.pilinfo(buf)
        out = buf.getvalue()
        lines = out.splitlines()
        self.assertEqual(lines[0], "-" * 68)
        self.assertTrue(lines[1].startswith("Pillow "))
        self.assertEqual(lines[2], "-" * 68)
        self.assertTrue(lines[3].startswith("Python modules loaded from "))
        self.assertTrue(lines[4].startswith("Binary modules loaded from "))
        self.assertEqual(lines[5], "-" * 68)
        self.assertTrue(lines[6].startswith("Python "))
        jpeg = (
            "\n"
            + "-" * 68
            + "\n"
            + "JPEG image/jpeg\n"
            + "Extensions: .jfif, .jpe, .jpeg, .jpg\n"
            + "Features: open, save\n"
            + "-" * 68
            + "\n"
        )
        self.assertIn(jpeg, out)



from PIL import Image, ImageEnhance

from .helper import PillowTestCase, hopper


class TestImageEnhance(PillowTestCase):
    def test_sanity(self):

        # FIXME: assert_image
        # Implicit asserts no exception:
        ImageEnhance.Color(hopper()).enhance(0.5)
        ImageEnhance.Contrast(hopper()).enhance(0.5)
        ImageEnhance.Brightness(hopper()).enhance(0.5)
        ImageEnhance.Sharpness(hopper()).enhance(0.5)

    def test_crash(self):

        # crashes on small images
        im = Image.new("RGB", (1, 1))
        ImageEnhance.Sharpness(im).enhance(0.5)

    def _half_transparent_image(self):
        # returns an image, half transparent, half solid
        im = hopper("RGB")

        transparent = Image.new("L", im.size, 0)
        solid = Image.new("L", (im.size[0] // 2, im.size[1]), 255)
        transparent.paste(solid, (0, 0))
        im.putalpha(transparent)

        return im

    def _check_alpha(self, im, original, op, amount):
        self.assertEqual(im.getbands(), original.getbands())
        self.assert_image_equal(
            im.getchannel("A"),
            original.getchannel("A"),
            "Diff on %s: %s" % (op, amount),
        )

    def test_alpha(self):
        # Issue https://github.com/python-pillow/Pillow/issues/899
        # Is alpha preserved through image enhancement?

        original = self._half_transparent_image()

        for op in ["Color", "Brightness", "Contrast", "Sharpness"]:
            for amount in [0, 0.5, 1.0]:
                self._check_alpha(
                    getattr(ImageEnhance, op)(original).enhance(amount),
                    original,
                    op,
                    amount,
                )



import array
import struct

from PIL import Image, ImagePath
from PIL._util import py3

from .helper import PillowTestCase


class TestImagePath(PillowTestCase):
    def test_path(self):

        p = ImagePath.Path(list(range(10)))

        # sequence interface
        self.assertEqual(len(p), 5)
        self.assertEqual(p[0], (0.0, 1.0))
        self.assertEqual(p[-1], (8.0, 9.0))
        self.assertEqual(list(p[:1]), [(0.0, 1.0)])
        with self.assertRaises(TypeError) as cm:
            p["foo"]
        self.assertEqual(str(cm.exception), "Path indices must be integers, not str")
        self.assertEqual(
            list(p), [(0.0, 1.0), (2.0, 3.0), (4.0, 5.0), (6.0, 7.0), (8.0, 9.0)]
        )

        # method sanity check
        self.assertEqual(
            p.tolist(), [(0.0, 1.0), (2.0, 3.0), (4.0, 5.0), (6.0, 7.0), (8.0, 9.0)]
        )
        self.assertEqual(
            p.tolist(1), [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
        )

        self.assertEqual(p.getbbox(), (0.0, 1.0, 8.0, 9.0))

        self.assertEqual(p.compact(5), 2)
        self.assertEqual(list(p), [(0.0, 1.0), (4.0, 5.0), (8.0, 9.0)])

        p.transform((1, 0, 1, 0, 1, 1))
        self.assertEqual(list(p), [(1.0, 2.0), (5.0, 6.0), (9.0, 10.0)])

        # alternative constructors
        p = ImagePath.Path([0, 1])
        self.assertEqual(list(p), [(0.0, 1.0)])
        p = ImagePath.Path([0.0, 1.0])
        self.assertEqual(list(p), [(0.0, 1.0)])
        p = ImagePath.Path([0, 1])
        self.assertEqual(list(p), [(0.0, 1.0)])
        p = ImagePath.Path([(0, 1)])
        self.assertEqual(list(p), [(0.0, 1.0)])
        p = ImagePath.Path(p)
        self.assertEqual(list(p), [(0.0, 1.0)])
        p = ImagePath.Path(p.tolist(0))
        self.assertEqual(list(p), [(0.0, 1.0)])
        p = ImagePath.Path(p.tolist(1))
        self.assertEqual(list(p), [(0.0, 1.0)])
        p = ImagePath.Path(array.array("f", [0, 1]))
        self.assertEqual(list(p), [(0.0, 1.0)])

        arr = array.array("f", [0, 1])
        if hasattr(arr, "tobytes"):
            p = ImagePath.Path(arr.tobytes())
        else:
            p = ImagePath.Path(arr.tostring())
        self.assertEqual(list(p), [(0.0, 1.0)])

    def test_overflow_segfault(self):
        # Some Pythons fail getting the argument as an integer, and it falls
        # through to the sequence. Seeing this on 32-bit Windows.
        with self.assertRaises((TypeError, MemoryError)):
            # post patch, this fails with a memory error
            x = evil()

            # This fails due to the invalid malloc above,
            # and segfaults
            for i in range(200000):
                if py3:
                    x[i] = b"0" * 16
                else:
                    x[i] = "0" * 16


class evil:
    def __init__(self):
        self.corrupt = Image.core.path(0x4000000000000000)

    def __getitem__(self, i):
        x = self.corrupt[i]
        return struct.pack("dd", x[0], x[1])

    def __setitem__(self, i, x):
        self.corrupt[i] = struct.unpack("dd", x)



from io import BytesIO

from PIL import Image, features

from .helper import PillowLeakTestCase, unittest

test_file = "Tests/images/hopper.webp"


@unittest.skipUnless(features.check("webp"), "WebP is not installed")
class TestWebPLeaks(PillowLeakTestCase):

    mem_limit = 3 * 1024  # kb
    iterations = 100

    def test_leak_load(self):
        with open(test_file, "rb") as f:
            im_data = f.read()

        def core():
            with Image.open(BytesIO(im_data)) as im:
                im.load()

        self._test_leak(core)



from PIL import __version__

from .helper import PillowTestCase, unittest

try:
    import pyroma
except ImportError:
    pyroma = None


@unittest.skipIf(pyroma is None, "Pyroma is not installed")
class TestPyroma(PillowTestCase):
    def test_pyroma(self):
        # Arrange
        data = pyroma.projectdata.get_data(".")

        # Act
        rating = pyroma.ratings.rate(data)

        # Assert
        if "rc" in __version__:
            # Pyroma needs to chill about RC versions and not kill all our tests.
            self.assertEqual(
                rating,
                (9, ["The package's version number does not comply with PEP-386."]),
            )

        else:
            # Should have a perfect score
            self.assertEqual(rating, (10, []))



from PIL import Image

from .helper import PillowTestCase, unittest


class TestLibImage(PillowTestCase):
    def test_setmode(self):

        im = Image.new("L", (1, 1), 255)
        im.im.setmode("1")
        self.assertEqual(im.im.getpixel((0, 0)), 255)
        im.im.setmode("L")
        self.assertEqual(im.im.getpixel((0, 0)), 255)

        im = Image.new("1", (1, 1), 1)
        im.im.setmode("L")
        self.assertEqual(im.im.getpixel((0, 0)), 255)
        im.im.setmode("1")
        self.assertEqual(im.im.getpixel((0, 0)), 255)

        im = Image.new("RGB", (1, 1), (1, 2, 3))
        im.im.setmode("RGB")
        self.assertEqual(im.im.getpixel((0, 0)), (1, 2, 3))
        im.im.setmode("RGBA")
        self.assertEqual(im.im.getpixel((0, 0)), (1, 2, 3, 255))
        im.im.setmode("RGBX")
        self.assertEqual(im.im.getpixel((0, 0)), (1, 2, 3, 255))
        im.im.setmode("RGB")
        self.assertEqual(im.im.getpixel((0, 0)), (1, 2, 3))

        self.assertRaises(ValueError, im.im.setmode, "L")
        self.assertRaises(ValueError, im.im.setmode, "RGBABCDE")


if __name__ == "__main__":
    unittest.main()



from PIL import Image

from .helper import PillowTestCase, hopper


class TestImageCrop(PillowTestCase):
    def test_crop(self):
        def crop(mode):
            im = hopper(mode)
            self.assert_image_equal(im.crop(), im)

            cropped = im.crop((50, 50, 100, 100))
            self.assertEqual(cropped.mode, mode)
            self.assertEqual(cropped.size, (50, 50))

        for mode in "1", "P", "L", "RGB", "I", "F":
            crop(mode)

    def test_wide_crop(self):
        def crop(*bbox):
            i = im.crop(bbox)
            h = i.histogram()
            while h and not h[-1]:
                del h[-1]
            return tuple(h)

        im = Image.new("L", (100, 100), 1)

        self.assertEqual(crop(0, 0, 100, 100), (0, 10000))
        self.assertEqual(crop(25, 25, 75, 75), (0, 2500))

        # sides
        self.assertEqual(crop(-25, 0, 25, 50), (1250, 1250))
        self.assertEqual(crop(0, -25, 50, 25), (1250, 1250))
        self.assertEqual(crop(75, 0, 125, 50), (1250, 1250))
        self.assertEqual(crop(0, 75, 50, 125), (1250, 1250))

        self.assertEqual(crop(-25, 25, 125, 75), (2500, 5000))
        self.assertEqual(crop(25, -25, 75, 125), (2500, 5000))

        # corners
        self.assertEqual(crop(-25, -25, 25, 25), (1875, 625))
        self.assertEqual(crop(75, -25, 125, 25), (1875, 625))
        self.assertEqual(crop(75, 75, 125, 125), (1875, 625))
        self.assertEqual(crop(-25, 75, 25, 125), (1875, 625))

    def test_negative_crop(self):
        # Check negative crop size (@PIL171)

        im = Image.new("L", (512, 512))
        im = im.crop((400, 400, 200, 200))

        self.assertEqual(im.size, (0, 0))
        self.assertEqual(len(im.getdata()), 0)
        self.assertRaises(IndexError, lambda: im.getdata()[0])

    def test_crop_float(self):
        # Check cropping floats are rounded to nearest integer
        # https://github.com/python-pillow/Pillow/issues/1744

        # Arrange
        im = Image.new("RGB", (10, 10))
        self.assertEqual(im.size, (10, 10))

        # Act
        cropped = im.crop((0.9, 1.1, 4.2, 5.8))

        # Assert
        self.assertEqual(cropped.size, (3, 5))

    def test_crop_crash(self):
        # Image.crop crashes prepatch with an access violation
        # apparently a use after free on windows, see
        # https://github.com/python-pillow/Pillow/issues/1077

        test_img = "Tests/images/bmp/g/pal8-0.bmp"
        extents = (1, 1, 10, 10)
        # works prepatch
        img = Image.open(test_img)
        img2 = img.crop(extents)
        img2.load()

        # fail prepatch
        img = Image.open(test_img)
        img = img.crop(extents)
        img.load()

    def test_crop_zero(self):

        im = Image.new("RGB", (0, 0), "white")

        cropped = im.crop((0, 0, 0, 0))
        self.assertEqual(cropped.size, (0, 0))

        cropped = im.crop((10, 10, 20, 20))
        self.assertEqual(cropped.size, (10, 10))
        self.assertEqual(cropped.getdata()[0], (0, 0, 0))

        im = Image.new("RGB", (0, 0))

        cropped = im.crop((10, 10, 20, 20))
        self.assertEqual(cropped.size, (10, 10))
        self.assertEqual(cropped.getdata()[2], (0, 0, 0))



from .helper import PillowTestCase, hopper


class TestUploader(PillowTestCase):
    def check_upload_equal(self):
        result = hopper("P").convert("RGB")
        target = hopper("RGB")
        self.assert_image_equal(result, target)

    def check_upload_similar(self):
        result = hopper("P").convert("RGB")
        target = hopper("RGB")
        self.assert_image_similar(result, target, 0)



import os
import sys
import urllib.parse
import urllib.request


def fetch(url):
    name = urllib.parse.urlsplit(url)[2].split("/")[-1]

    if not os.path.exists(name):
        print("Fetching", url)
        try:
            r = urllib.request.urlopen(url)
        except urllib.error.URLError:
            r = urllib.request.urlopen(url)
        content = r.read()
        with open(name, "wb") as fd:
            fd.write(content)
    return name


if __name__ == "__main__":
    fetch(sys.argv[1])



from __future__ import print_function

from PIL import Image

from .helper import PillowTestCase, hopper, unittest

try:
    import numpy
except ImportError:
    numpy = None


TEST_IMAGE_SIZE = (10, 10)


@unittest.skipIf(numpy is None, "Numpy is not installed")
class TestNumpy(PillowTestCase):
    def test_numpy_to_image(self):
        def to_image(dtype, bands=1, boolean=0):
            if bands == 1:
                if boolean:
                    data = [0, 255] * 50
                else:
                    data = list(range(100))
                a = numpy.array(data, dtype=dtype)
                a.shape = TEST_IMAGE_SIZE
                i = Image.fromarray(a)
                if list(i.getdata()) != data:
                    print("data mismatch for", dtype)
            else:
                data = list(range(100))
                a = numpy.array([[x] * bands for x in data], dtype=dtype)
                a.shape = TEST_IMAGE_SIZE[0], TEST_IMAGE_SIZE[1], bands
                i = Image.fromarray(a)
                if list(i.getchannel(0).getdata()) != list(range(100)):
                    print("data mismatch for", dtype)
            return i

        # Check supported 1-bit integer formats
        self.assert_image(to_image(numpy.bool, 1, 1), "1", TEST_IMAGE_SIZE)
        self.assert_image(to_image(numpy.bool8, 1, 1), "1", TEST_IMAGE_SIZE)

        # Check supported 8-bit integer formats
        self.assert_image(to_image(numpy.uint8), "L", TEST_IMAGE_SIZE)
        self.assert_image(to_image(numpy.uint8, 3), "RGB", TEST_IMAGE_SIZE)
        self.assert_image(to_image(numpy.uint8, 4), "RGBA", TEST_IMAGE_SIZE)
        self.assert_image(to_image(numpy.int8), "I", TEST_IMAGE_SIZE)

        # Check non-fixed-size integer types
        # These may fail, depending on the platform, since we have no native
        # 64 bit int image types.
        # self.assert_image(to_image(numpy.uint), "I", TEST_IMAGE_SIZE)
        # self.assert_image(to_image(numpy.int), "I", TEST_IMAGE_SIZE)

        # Check 16-bit integer formats
        if Image._ENDIAN == "<":
            self.assert_image(to_image(numpy.uint16), "I;16", TEST_IMAGE_SIZE)
        else:
            self.assert_image(to_image(numpy.uint16), "I;16B", TEST_IMAGE_SIZE)

        self.assert_image(to_image(numpy.int16), "I", TEST_IMAGE_SIZE)

        # Check 32-bit integer formats
        self.assert_image(to_image(numpy.uint32), "I", TEST_IMAGE_SIZE)
        self.assert_image(to_image(numpy.int32), "I", TEST_IMAGE_SIZE)

        # Check 64-bit integer formats
        self.assertRaises(TypeError, to_image, numpy.uint64)
        self.assertRaises(TypeError, to_image, numpy.int64)

        # Check floating-point formats
        self.assert_image(to_image(numpy.float), "F", TEST_IMAGE_SIZE)
        self.assertRaises(TypeError, to_image, numpy.float16)
        self.assert_image(to_image(numpy.float32), "F", TEST_IMAGE_SIZE)
        self.assert_image(to_image(numpy.float64), "F", TEST_IMAGE_SIZE)

        self.assert_image(to_image(numpy.uint8, 2), "LA", (10, 10))
        self.assert_image(to_image(numpy.uint8, 3), "RGB", (10, 10))
        self.assert_image(to_image(numpy.uint8, 4), "RGBA", (10, 10))

    # based on an erring example at
    # https://stackoverflow.com/questions/10854903/what-is-causing-dimension-dependent-attributeerror-in-pil-fromarray-function
    def test_3d_array(self):
        size = (5, TEST_IMAGE_SIZE[0], TEST_IMAGE_SIZE[1])
        a = numpy.ones(size, dtype=numpy.uint8)
        self.assert_image(Image.fromarray(a[1, :, :]), "L", TEST_IMAGE_SIZE)
        size = (TEST_IMAGE_SIZE[0], 5, TEST_IMAGE_SIZE[1])
        a = numpy.ones(size, dtype=numpy.uint8)
        self.assert_image(Image.fromarray(a[:, 1, :]), "L", TEST_IMAGE_SIZE)
        size = (TEST_IMAGE_SIZE[0], TEST_IMAGE_SIZE[1], 5)
        a = numpy.ones(size, dtype=numpy.uint8)
        self.assert_image(Image.fromarray(a[:, :, 1]), "L", TEST_IMAGE_SIZE)

    def _test_img_equals_nparray(self, img, np):
        self.assertGreaterEqual(len(np.shape), 2)
        np_size = np.shape[1], np.shape[0]
        self.assertEqual(img.size, np_size)
        px = img.load()
        for x in range(0, img.size[0], int(img.size[0] / 10)):
            for y in range(0, img.size[1], int(img.size[1] / 10)):
                self.assert_deep_equal(px[x, y], np[y, x])

    def test_16bit(self):
        img = Image.open("Tests/images/16bit.cropped.tif")
        np_img = numpy.array(img)
        self._test_img_equals_nparray(img, np_img)
        self.assertEqual(np_img.dtype, numpy.dtype("<u2"))

    def test_1bit(self):
        # Test that 1-bit arrays convert to numpy and back
        # See: https://github.com/python-pillow/Pillow/issues/350
        arr = numpy.array([[1, 0, 0, 1, 0], [0, 1, 0, 0, 0]], "u1")
        img = Image.fromarray(arr * 255).convert("1")
        self.assertEqual(img.mode, "1")
        arr_back = numpy.array(img)
        numpy.testing.assert_array_equal(arr, arr_back)

    def test_save_tiff_uint16(self):
        # Tests that we're getting the pixel value in the right byte order.
        pixel_value = 0x1234
        a = numpy.array(
            [pixel_value] * TEST_IMAGE_SIZE[0] * TEST_IMAGE_SIZE[1], dtype=numpy.uint16
        )
        a.shape = TEST_IMAGE_SIZE
        img = Image.fromarray(a)

        img_px = img.load()
        self.assertEqual(img_px[0, 0], pixel_value)

    def test_to_array(self):
        def _to_array(mode, dtype):
            img = hopper(mode)

            # Resize to non-square
            img = img.crop((3, 0, 124, 127))
            self.assertEqual(img.size, (121, 127))

            np_img = numpy.array(img)
            self._test_img_equals_nparray(img, np_img)
            self.assertEqual(np_img.dtype, dtype)

        modes = [
            ("L", numpy.uint8),
            ("I", numpy.int32),
            ("F", numpy.float32),
            ("LA", numpy.uint8),
            ("RGB", numpy.uint8),
            ("RGBA", numpy.uint8),
            ("RGBX", numpy.uint8),
            ("CMYK", numpy.uint8),
            ("YCbCr", numpy.uint8),
            ("I;16", "<u2"),
            ("I;16B", ">u2"),
            ("I;16L", "<u2"),
            ("HSV", numpy.uint8),
        ]

        for mode in modes:
            _to_array(*mode)

    def test_point_lut(self):
        # see https://github.com/python-pillow/Pillow/issues/439

        data = list(range(256)) * 3
        lut = numpy.array(data, dtype=numpy.uint8)

        im = hopper()

        im.point(lut)

    def test_putdata(self):
        # shouldn't segfault
        # see https://github.com/python-pillow/Pillow/issues/1008

        im = Image.new("F", (150, 100))
        arr = numpy.zeros((15000,), numpy.float32)
        im.putdata(arr)

        self.assertEqual(len(im.getdata()), len(arr))

    def test_roundtrip_eye(self):
        for dtype in (
            numpy.bool,
            numpy.bool8,
            numpy.int8,
            numpy.int16,
            numpy.int32,
            numpy.uint8,
            numpy.uint16,
            numpy.uint32,
            numpy.float,
            numpy.float32,
            numpy.float64,
        ):
            arr = numpy.eye(10, dtype=dtype)
            numpy.testing.assert_array_equal(arr, numpy.array(Image.fromarray(arr)))

    def test_zero_size(self):
        # Shouldn't cause floating point exception
        # See https://github.com/python-pillow/Pillow/issues/2259

        im = Image.fromarray(numpy.empty((0, 0), dtype=numpy.uint8))

        self.assertEqual(im.size, (0, 0))

    def test_bool(self):
        # https://github.com/python-pillow/Pillow/issues/2044
        a = numpy.zeros((10, 2), dtype=numpy.bool)
        a[0][0] = True

        im2 = Image.fromarray(a)
        self.assertEqual(im2.getdata()[0], 255)

    def test_no_resource_warning_for_numpy_array(self):
        # https://github.com/python-pillow/Pillow/issues/835
        # Arrange
        from numpy import array

        test_file = "Tests/images/hopper.png"
        im = Image.open(test_file)

        # Act/Assert
        self.assert_warning(None, lambda: array(im))



from PIL import Image

from .helper import PillowTestCase, hopper


class TestImageQuantize(PillowTestCase):
    def test_sanity(self):
        image = hopper()
        converted = image.quantize()
        self.assert_image(converted, "P", converted.size)
        self.assert_image_similar(converted.convert("RGB"), image, 10)

        image = hopper()
        converted = image.quantize(palette=hopper("P"))
        self.assert_image(converted, "P", converted.size)
        self.assert_image_similar(converted.convert("RGB"), image, 60)

    def test_libimagequant_quantize(self):
        image = hopper()
        try:
            converted = image.quantize(100, Image.LIBIMAGEQUANT)
        except ValueError as ex:
            if "dependency" in str(ex).lower():
                self.skipTest("libimagequant support not available")
            else:
                raise
        self.assert_image(converted, "P", converted.size)
        self.assert_image_similar(converted.convert("RGB"), image, 15)
        self.assertEqual(len(converted.getcolors()), 100)

    def test_octree_quantize(self):
        image = hopper()
        converted = image.quantize(100, Image.FASTOCTREE)
        self.assert_image(converted, "P", converted.size)
        self.assert_image_similar(converted.convert("RGB"), image, 20)
        self.assertEqual(len(converted.getcolors()), 100)

    def test_rgba_quantize(self):
        image = hopper("RGBA")
        self.assertRaises(ValueError, image.quantize, method=0)

        self.assertEqual(image.quantize().convert().mode, "RGBA")

    def test_quantize(self):
        image = Image.open("Tests/images/caption_6_33_22.png").convert("RGB")
        converted = image.quantize()
        self.assert_image(converted, "P", converted.size)
        self.assert_image_similar(converted.convert("RGB"), image, 1)

    def test_quantize_no_dither(self):
        image = hopper()
        palette = Image.open("Tests/images/caption_6_33_22.png").convert("P")

        converted = image.quantize(dither=0, palette=palette)
        self.assert_image(converted, "P", converted.size)

    def test_quantize_dither_diff(self):
        image = hopper()
        palette = Image.open("Tests/images/caption_6_33_22.png").convert("P")

        dither = image.quantize(dither=1, palette=palette)
        nodither = image.quantize(dither=0, palette=palette)

        self.assertNotEqual(dither.tobytes(), nodither.tobytes())



from PIL import FontFile, Image, ImageDraw, ImageFont, PcfFontFile
from PIL._util import py3

from .helper import PillowTestCase

codecs = dir(Image.core)

fontname = "Tests/fonts/10x20-ISO8859-1.pcf"

message = "hello, world"


class TestFontPcf(PillowTestCase):
    def setUp(self):
        if "zip_encoder" not in codecs or "zip_decoder" not in codecs:
            self.skipTest("zlib support not available")

    def save_font(self):
        with open(fontname, "rb") as test_file:
            font = PcfFontFile.PcfFontFile(test_file)
        self.assertIsInstance(font, FontFile.FontFile)
        # check the number of characters in the font
        self.assertEqual(len([_f for _f in font.glyph if _f]), 223)

        tempname = self.tempfile("temp.pil")
        self.addCleanup(self.delete_tempfile, tempname[:-4] + ".pbm")
        font.save(tempname)

        with Image.open(tempname.replace(".pil", ".pbm")) as loaded:
            with Image.open("Tests/fonts/10x20.pbm") as target:
                self.assert_image_equal(loaded, target)

        with open(tempname, "rb") as f_loaded:
            with open("Tests/fonts/10x20.pil", "rb") as f_target:
                self.assertEqual(f_loaded.read(), f_target.read())
        return tempname

    def test_sanity(self):
        self.save_font()

    def test_invalid_file(self):
        with open("Tests/images/flower.jpg", "rb") as fp:
            self.assertRaises(SyntaxError, PcfFontFile.PcfFontFile, fp)

    def test_draw(self):
        tempname = self.save_font()
        font = ImageFont.load(tempname)
        im = Image.new("L", (130, 30), "white")
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), message, "black", font=font)
        with Image.open("Tests/images/test_draw_pbm_target.png") as target:
            self.assert_image_similar(im, target, 0)

    def test_textsize(self):
        tempname = self.save_font()
        font = ImageFont.load(tempname)
        for i in range(255):
            (dx, dy) = font.getsize(chr(i))
            self.assertEqual(dy, 20)
            self.assertIn(dx, (0, 10))
        for l in range(len(message)):
            msg = message[: l + 1]
            self.assertEqual(font.getsize(msg), (len(msg) * 10, 20))

    def _test_high_characters(self, message):
        tempname = self.save_font()
        font = ImageFont.load(tempname)
        im = Image.new("L", (750, 30), "white")
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), message, "black", font=font)
        with Image.open("Tests/images/high_ascii_chars.png") as target:
            self.assert_image_similar(im, target, 0)

    def test_high_characters(self):
        message = "".join(chr(i + 1) for i in range(140, 232))
        self._test_high_characters(message)
        # accept bytes instances in Py3.
        if py3:
            self._test_high_characters(message.encode("latin1"))



import os

from fetch import fetch

if __name__ == "__main__":
    for version in ["2.7.15", "3.4.4"]:
        for platform in ["", ".amd64"]:
            for extension in ["", ".asc"]:
                fetch(
                    "https://www.python.org/ftp/python/%s/python-%s%s.msi%s"
                    % (version, version, platform, extension)
                )

    # find pip, if it's not in the path!
    os.system("pip install virtualenv")



import io
import os
import os.path
import tempfile
import time

from PIL import Image, PdfParser

from .helper import PillowTestCase, hopper


class TestFilePdf(PillowTestCase):
    def helper_save_as_pdf(self, mode, **kwargs):
        # Arrange
        im = hopper(mode)
        outfile = self.tempfile("temp_" + mode + ".pdf")

        # Act
        im.save(outfile, **kwargs)

        # Assert
        self.assertTrue(os.path.isfile(outfile))
        self.assertGreater(os.path.getsize(outfile), 0)
        with PdfParser.PdfParser(outfile) as pdf:
            if kwargs.get("append_images", False) or kwargs.get("append", False):
                self.assertGreater(len(pdf.pages), 1)
            else:
                self.assertGreater(len(pdf.pages), 0)
        with open(outfile, "rb") as fp:
            contents = fp.read()
        size = tuple(
            int(d)
            for d in contents.split(b"/MediaBox [ 0 0 ")[1].split(b"]")[0].split()
        )
        self.assertEqual(im.size, size)

        return outfile

    def test_monochrome(self):
        # Arrange
        mode = "1"

        # Act / Assert
        self.helper_save_as_pdf(mode)

    def test_greyscale(self):
        # Arrange
        mode = "L"

        # Act / Assert
        self.helper_save_as_pdf(mode)

    def test_rgb(self):
        # Arrange
        mode = "RGB"

        # Act / Assert
        self.helper_save_as_pdf(mode)

    def test_p_mode(self):
        # Arrange
        mode = "P"

        # Act / Assert
        self.helper_save_as_pdf(mode)

    def test_cmyk_mode(self):
        # Arrange
        mode = "CMYK"

        # Act / Assert
        self.helper_save_as_pdf(mode)

    def test_unsupported_mode(self):
        im = hopper("LA")
        outfile = self.tempfile("temp_LA.pdf")

        self.assertRaises(ValueError, im.save, outfile)

    def test_save_all(self):
        # Single frame image
        self.helper_save_as_pdf("RGB", save_all=True)

        # Multiframe image
        im = Image.open("Tests/images/dispose_bgnd.gif")

        outfile = self.tempfile("temp.pdf")
        im.save(outfile, save_all=True)

        self.assertTrue(os.path.isfile(outfile))
        self.assertGreater(os.path.getsize(outfile), 0)

        # Append images
        ims = [hopper()]
        im.copy().save(outfile, save_all=True, append_images=ims)

        self.assertTrue(os.path.isfile(outfile))
        self.assertGreater(os.path.getsize(outfile), 0)

        # Test appending using a generator
        def imGenerator(ims):
            for im in ims:
                yield im

        im.save(outfile, save_all=True, append_images=imGenerator(ims))

        self.assertTrue(os.path.isfile(outfile))
        self.assertGreater(os.path.getsize(outfile), 0)

        # Append JPEG images
        jpeg = Image.open("Tests/images/flower.jpg")
        jpeg.save(outfile, save_all=True, append_images=[jpeg.copy()])

        self.assertTrue(os.path.isfile(outfile))
        self.assertGreater(os.path.getsize(outfile), 0)

    def test_multiframe_normal_save(self):
        # Test saving a multiframe image without save_all
        im = Image.open("Tests/images/dispose_bgnd.gif")

        outfile = self.tempfile("temp.pdf")
        im.save(outfile)

        self.assertTrue(os.path.isfile(outfile))
        self.assertGreater(os.path.getsize(outfile), 0)

    def test_pdf_open(self):
        # fail on a buffer full of null bytes
        self.assertRaises(
            PdfParser.PdfFormatError, PdfParser.PdfParser, buf=bytearray(65536)
        )

        # make an empty PDF object
        with PdfParser.PdfParser() as empty_pdf:
            self.assertEqual(len(empty_pdf.pages), 0)
            self.assertEqual(len(empty_pdf.info), 0)
            self.assertFalse(empty_pdf.should_close_buf)
            self.assertFalse(empty_pdf.should_close_file)

        # make a PDF file
        pdf_filename = self.helper_save_as_pdf("RGB")

        # open the PDF file
        with PdfParser.PdfParser(filename=pdf_filename) as hopper_pdf:
            self.assertEqual(len(hopper_pdf.pages), 1)
            self.assertTrue(hopper_pdf.should_close_buf)
            self.assertTrue(hopper_pdf.should_close_file)

        # read a PDF file from a buffer with a non-zero offset
        with open(pdf_filename, "rb") as f:
            content = b"xyzzy" + f.read()
        with PdfParser.PdfParser(buf=content, start_offset=5) as hopper_pdf:
            self.assertEqual(len(hopper_pdf.pages), 1)
            self.assertFalse(hopper_pdf.should_close_buf)
            self.assertFalse(hopper_pdf.should_close_file)

        # read a PDF file from an already open file
        with open(pdf_filename, "rb") as f:
            with PdfParser.PdfParser(f=f) as hopper_pdf:
                self.assertEqual(len(hopper_pdf.pages), 1)
                self.assertTrue(hopper_pdf.should_close_buf)
                self.assertFalse(hopper_pdf.should_close_file)

    def test_pdf_append_fails_on_nonexistent_file(self):
        im = hopper("RGB")
        temp_dir = tempfile.mkdtemp()
        try:
            self.assertRaises(
                IOError, im.save, os.path.join(temp_dir, "nonexistent.pdf"), append=True
            )
        finally:
            os.rmdir(temp_dir)

    def check_pdf_pages_consistency(self, pdf):
        pages_info = pdf.read_indirect(pdf.pages_ref)
        self.assertNotIn(b"Parent", pages_info)
        self.assertIn(b"Kids", pages_info)
        kids_not_used = pages_info[b"Kids"]
        for page_ref in pdf.pages:
            while True:
                if page_ref in kids_not_used:
                    kids_not_used.remove(page_ref)
                page_info = pdf.read_indirect(page_ref)
                self.assertIn(b"Parent", page_info)
                page_ref = page_info[b"Parent"]
                if page_ref == pdf.pages_ref:
                    break
            self.assertEqual(pdf.pages_ref, page_info[b"Parent"])
        self.assertEqual(kids_not_used, [])

    def test_pdf_append(self):
        # make a PDF file
        pdf_filename = self.helper_save_as_pdf("RGB", producer="PdfParser")

        # open it, check pages and info
        with PdfParser.PdfParser(pdf_filename, mode="r+b") as pdf:
            self.assertEqual(len(pdf.pages), 1)
            self.assertEqual(len(pdf.info), 4)
            self.assertEqual(
                pdf.info.Title, os.path.splitext(os.path.basename(pdf_filename))[0]
            )
            self.assertEqual(pdf.info.Producer, "PdfParser")
            self.assertIn(b"CreationDate", pdf.info)
            self.assertIn(b"ModDate", pdf.info)
            self.check_pdf_pages_consistency(pdf)

            # append some info
            pdf.info.Title = "abc"
            pdf.info.Author = "def"
            pdf.info.Subject = u"ghi\uABCD"
            pdf.info.Keywords = "qw)e\\r(ty"
            pdf.info.Creator = "hopper()"
            pdf.start_writing()
            pdf.write_xref_and_trailer()

        # open it again, check pages and info again
        with PdfParser.PdfParser(pdf_filename) as pdf:
            self.assertEqual(len(pdf.pages), 1)
            self.assertEqual(len(pdf.info), 8)
            self.assertEqual(pdf.info.Title, "abc")
            self.assertIn(b"CreationDate", pdf.info)
            self.assertIn(b"ModDate", pdf.info)
            self.check_pdf_pages_consistency(pdf)

        # append two images
        mode_CMYK = hopper("CMYK")
        mode_P = hopper("P")
        mode_CMYK.save(pdf_filename, append=True, save_all=True, append_images=[mode_P])

        # open the PDF again, check pages and info again
        with PdfParser.PdfParser(pdf_filename) as pdf:
            self.assertEqual(len(pdf.pages), 3)
            self.assertEqual(len(pdf.info), 8)
            self.assertEqual(PdfParser.decode_text(pdf.info[b"Title"]), "abc")
            self.assertEqual(pdf.info.Title, "abc")
            self.assertEqual(pdf.info.Producer, "PdfParser")
            self.assertEqual(pdf.info.Keywords, "qw)e\\r(ty")
            self.assertEqual(pdf.info.Subject, u"ghi\uABCD")
            self.assertIn(b"CreationDate", pdf.info)
            self.assertIn(b"ModDate", pdf.info)
            self.check_pdf_pages_consistency(pdf)

    def test_pdf_info(self):
        # make a PDF file
        pdf_filename = self.helper_save_as_pdf(
            "RGB",
            title="title",
            author="author",
            subject="subject",
            keywords="keywords",
            creator="creator",
            producer="producer",
            creationDate=time.strptime("2000", "%Y"),
            modDate=time.strptime("2001", "%Y"),
        )

        # open it, check pages and info
        with PdfParser.PdfParser(pdf_filename) as pdf:
            self.assertEqual(len(pdf.info), 8)
            self.assertEqual(pdf.info.Title, "title")
            self.assertEqual(pdf.info.Author, "author")
            self.assertEqual(pdf.info.Subject, "subject")
            self.assertEqual(pdf.info.Keywords, "keywords")
            self.assertEqual(pdf.info.Creator, "creator")
            self.assertEqual(pdf.info.Producer, "producer")
            self.assertEqual(pdf.info.CreationDate, time.strptime("2000", "%Y"))
            self.assertEqual(pdf.info.ModDate, time.strptime("2001", "%Y"))
            self.check_pdf_pages_consistency(pdf)

    def test_pdf_append_to_bytesio(self):
        im = hopper("RGB")
        f = io.BytesIO()
        im.save(f, format="PDF")
        initial_size = len(f.getvalue())
        self.assertGreater(initial_size, 0)
        im = hopper("P")
        f = io.BytesIO(f.getvalue())
        im.save(f, format="PDF", append=True)
        self.assertGreater(len(f.getvalue()), initial_size)



import os
import sys

from PIL import Image, PSDraw

from .helper import PillowTestCase


class TestPsDraw(PillowTestCase):
    def _create_document(self, ps):
        im = Image.open("Tests/images/hopper.ppm")
        title = "hopper"
        box = (1 * 72, 2 * 72, 7 * 72, 10 * 72)  # in points

        ps.begin_document(title)

        # draw diagonal lines in a cross
        ps.line((1 * 72, 2 * 72), (7 * 72, 10 * 72))
        ps.line((7 * 72, 2 * 72), (1 * 72, 10 * 72))

        # draw the image (75 dpi)
        ps.image(box, im, 75)
        ps.rectangle(box)

        # draw title
        ps.setfont("Courier", 36)
        ps.text((3 * 72, 4 * 72), title)

        ps.end_document()

    def test_draw_postscript(self):

        # Based on Pillow tutorial, but there is no textsize:
        # https://pillow.readthedocs.io/en/latest/handbook/tutorial.html#drawing-postscript

        # Arrange
        tempfile = self.tempfile("temp.ps")
        with open(tempfile, "wb") as fp:
            # Act
            ps = PSDraw.PSDraw(fp)
            self._create_document(ps)

        # Assert
        # Check non-zero file was created
        self.assertTrue(os.path.isfile(tempfile))
        self.assertGreater(os.path.getsize(tempfile), 0)

    def test_stdout(self):
        # Temporarily redirect stdout
        try:
            from cStringIO import StringIO
        except ImportError:
            from io import StringIO
        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()

        ps = PSDraw.PSDraw()
        self._create_document(ps)

        # Reset stdout
        sys.stdout = old_stdout

        self.assertNotEqual(mystdout.getvalue(), "")
